# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import builtins

VERSION_REQURE_FILE = 3
ENV_STORAGE_FILE = "RESPONSE_FILE"
ENV_REPLACEMENT_FILE = "REPLACEMENT_FILE"
ENV_REPLACEMENT_NAME = "REPLACEMENT_VAR"
ENV_DEBUG = "DEBUG"
ENV_APPLY_LATENCY = "LATENCY"
ENV_REQURE_STORAGE_MODE = "REQURE_MODE"
REPLACE_DEFAULT_KEY = "FILTERS"
METATADA_KEY = "_requre"
KEY_MINIMAL_MATCH = 2
RELATIVE_TEST_DATA_DIRECTORY = "test_data"
DEFAULT_SUFIX = "yaml"
REQURE_CASSETTE_ATTRIBUTE_NAME = "_requre_cassette"
REQURE_SETUP_APPLIED_ATTRIBUTE_NAME = "_requre_cassette_setup_applied"
TEST_METHOD_REGEXP = "test.*"

DEFAULT_IMPORT_FUNCTION = builtins.__import__
